"use strict";
/**
 * `IxState` (Indexed State) is a derivation of `State`, that may have **polymorphic** states;
 * The input may be of a different type than the output.
 *
 * State changes are kept track of by the generics `I`, `O`, `X` and `Z`.
 * - `I` indicates the current input state
 * - `O` indicates the current output state
 * - `Z` indicates the next input state
 * - `Z` indicates the next output state
 *
 * When composing `IxState` where the indexes are polymorhpic, functions `ixchain` and `ixap` can be
 * used to compose `<I, O>` with `<O, Z>` to return `<I, Z>`.
 *
 * Intuitively this is function composition on the indexes.
 *
 *
 * `IxState` derived by applying the `Identity` monad to the transformer `IxStateT`.
 * @since 2.10.0
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.apS = exports.ixapS = exports.discard = exports.ixdiscard = exports.bindTo = exports.bind = exports.ixbindTo = exports.ixbind = exports.let = exports.let_ = exports.ixlet = exports.Do = exports.ixDo = exports.local = exports.chain = exports.ap = exports.flap = exports.map = exports.of = exports.chainFirst = exports.ixchainFirst = exports.ixchain = exports.apSecond = exports.apFirst = exports.ixapSecond = exports.ixapFirst = exports.ixap = exports.ixflap = exports.ixmap = exports.ixof = exports.Monad = exports.Chain = exports.Apply = exports.Functor = exports.Pointed = exports.IxMonad = exports.IxApplicative = exports.IxPointed = exports.IxChain = exports.IxApply = exports.IxFunctor = exports.evaluate = exports.ixexecute = exports.toState = exports.gets = exports.get = exports.fromState = exports.put = exports.ixmodify = exports.URI = void 0;
var Apply_1 = require("fp-ts/Apply");
var Chain_1 = require("fp-ts/Chain");
var Functor_1 = require("fp-ts/Functor");
var I = __importStar(require("fp-ts/Identity"));
var IxApply_1 = require("./IxApply");
var IxChain_1 = require("./IxChain");
var IxFunctor_1 = require("./IxFunctor");
var IxStateT = __importStar(require("./IxStateT"));
/**
 * @category Model
 * @since 2.10.0
 */
exports.URI = 'IxState';
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @summary
 * Apply a function to the state, polymorphically changing the resulting state.
 *
 * @category constructors
 * @since 2.10.0
 */
exports.ixmodify = 
/*#__PURE__*/
IxStateT.ixmodify(I.Pointed);
/**
 * @category constructors
 * @since 2.10.0
 */
exports.put = 
/*#__PURE__*/
IxStateT.put(I.Pointed);
/**
 * @category constructors
 * @since 2.10.0
 */
exports.fromState = IxStateT.fromStateF();
/**
 * @category constructors
 * @since 2.10.0
 */
exports.get = 
/*#__PURE__*/
IxStateT.get(I.Pointed);
/**
 * @category constructors
 * @since 2.10.0
 */
exports.gets = 
/*#__PURE__*/
IxStateT.gets(I.Pointed);
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * @category destructors
 * @since 2.10.0
 */
exports.toState = IxStateT.toStateF();
/**
 * @category destructors
 * @since 2.10.0
 */
exports.ixexecute = 
/*#__PURE__*/
IxStateT.ixexecute(I.Functor);
/**
 * @category destructors
 * @since 2.10.0
 */
exports.evaluate = 
/*#__PURE__*/
IxStateT.evaluate(I.Functor);
// -------------------------------------------------------------------------------------
// non-pipeables
// -------------------------------------------------------------------------------------
var _ixof = function (a) { return function (i) { return [a, i]; }; };
var _ixmap = function (fa, f) { return function (i) {
    var _a = fa(i), a = _a[0], o = _a[1];
    return [f(a), o];
}; };
var _ixap = function (fab, fa) { return function (i) {
    var _a = fab(i), f = _a[0], o = _a[1];
    var _b = fa(o), a = _b[0], z = _b[1];
    return [f(a), z];
}; };
var _ixchain = function (fa, f) { return function (i) {
    var _a = fa(i), a = _a[0], o = _a[1];
    var _b = f(a)(o), b = _b[0], z = _b[1];
    return [b, z];
}; };
var _of = _ixof;
var _map = _ixmap;
var _ap = _ixap;
var _chain = _ixchain;
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 2.10.0
 */
exports.IxFunctor = {
    URI: exports.URI,
    ixmap: function (f) { return function (fa) { return _ixmap(fa, f); }; },
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.IxApply = {
    URI: exports.URI,
    ixmap: exports.IxFunctor.ixmap,
    ixap: function (fa) { return function (fab) { return _ixap(fab, fa); }; },
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.IxChain = {
    URI: exports.URI,
    ixmap: exports.IxFunctor.ixmap,
    ixap: exports.IxApply.ixap,
    ixchain: function (f) { return function (fa) { return _ixchain(fa, f); }; },
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.IxPointed = {
    URI: exports.URI,
    ixof: _ixof,
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.IxApplicative = {
    URI: exports.URI,
    ixof: exports.IxPointed.ixof,
    ixmap: exports.IxFunctor.ixmap,
    ixap: exports.IxApply.ixap,
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.IxMonad = {
    URI: exports.URI,
    ixof: exports.IxPointed.ixof,
    ixmap: exports.IxFunctor.ixmap,
    ixap: exports.IxApply.ixap,
    ixchain: exports.IxChain.ixchain,
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.Pointed = { URI: exports.URI, of: _of };
/**
 * @category instances
 * @since 2.10.0
 */
exports.Functor = {
    URI: exports.URI,
    map: _ixmap,
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.Apply = {
    URI: exports.URI,
    map: _map,
    ap: _ap,
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.Chain = {
    URI: exports.URI,
    map: _map,
    ap: _ap,
    chain: _chain,
};
/**
 * @category instances
 * @since 2.10.0
 */
exports.Monad = {
    URI: exports.URI,
    of: _of,
    map: _map,
    ap: _ap,
    chain: _chain,
};
// -------------------------------------------------------------------------------------
// type class members
// -------------------------------------------------------------------------------------
/**
 * @category IxPointed
 * @since 2.10.0
 */
exports.ixof = _ixof;
/**
 * @category IxFunctor
 * @since 2.10.0
 */
exports.ixmap = 
/*#__PURE__*/
IxStateT.ixmap(I.Functor);
/**
 * @category IxFunctor
 * @since 2.10.0
 */
exports.ixflap = 
/*#__PURE__*/
IxFunctor_1.ixflap(exports.IxFunctor);
/**
 * Apply a function to an argument under a type constructor.
 *
 * @category IxApply
 * @since 2.10.0
 */
exports.ixap = 
/*#__PURE__*/
IxStateT.ixap(I.Chain);
/**
 * @category IxApply
 *
 * @since 2.10.0
 */
exports.ixapFirst = 
/*#__PURE__*/
IxApply_1.ixapFirst(exports.IxApply);
/**
 * @category IxApply
 * @since 2.10.0
 */
exports.ixapSecond = 
/*#__PURE__*/
IxApply_1.ixapSecond(exports.IxApply);
/**
 * @category Apply
 *
 * @since 2.10.0
 */
exports.apFirst = 
/*#__PURE__*/
Apply_1.apFirst(exports.Apply);
/**
 * @category Apply
 * @since 2.10.0
 */
exports.apSecond = 
/*#__PURE__*/
Apply_1.apSecond(exports.Apply);
/**
 * Composes computations in sequence, using the return value of one computation to determine the next computation.
 *
 * @category IxChain
 *
 * @since 2.10.0
 */
exports.ixchain = 
/*#__PURE__*/
IxStateT.ixchain(I.Chain);
/**
 * @category IxChain
 * @since 2.10.0
 */
exports.ixchainFirst = 
/*#__PURE__*/
IxChain_1.ixchainFirst(exports.IxChain);
/**
 * @category Chain
 * @since 2.10.0
 */
exports.chainFirst = 
/*#__PURE__*/
Chain_1.chainFirst(exports.Chain);
/**
 * @category Pointed
 * @since 2.10.0
 */
exports.of = 
/*#__PURE__*/
IxStateT.of(I.Pointed);
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 2.10.0
 */
exports.map = 
/*#__PURE__*/
IxStateT.map(I.Functor);
/**
 * `map` can be used to turn functions `(a: A) => B` into functions `(fa: F<A>) => F<B>` whose argument and return types
 * use the type constructor `F` to represent some computational context.
 *
 * @category Functor
 * @since 2.10.0
 */
exports.flap = 
/*#__PURE__*/
Functor_1.flap(exports.Functor);
/**
 *
 * @category Apply
 * @since 2.10.0
 */
exports.ap = 
/*#__PURE__*/
IxStateT.ap(I.Chain);
/**
 *
 * @category Chain
 * @since 2.10.0
 */
exports.chain = 
/*#__PURE__*/
IxStateT.chain(I.Chain);
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * Changes the value of the local context during the execution of the action `ma` (similar to `Contravariant`'s
 * `contramap`).
 *
 * @category combinators
 * @since 2.10.0
 */
exports.local = IxStateT.local();
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @since 2.10.0
 */
exports.ixDo = 
/*#__PURE__*/
IxChain_1.ixDo(exports.Pointed);
/**
 * @since 2.10.0
 */
exports.Do = exports.ixDo;
/**
 * @since 2.10.2
 */
exports.ixlet = 
/*#__PURE__*/
IxFunctor_1.ixlet(exports.IxFunctor);
/**
 * @since 2.10.2
 */
exports.let_ = exports.ixlet;
exports.let = exports.let_;
/**
 * @since 2.10.0
 */
exports.ixbind = 
/*#__PURE__*/
IxChain_1.ixbind(exports.IxChain);
/**
 * @since 2.10.0
 */
exports.ixbindTo = 
/*#__PURE__*/
IxFunctor_1.ixbindTo(exports.IxFunctor);
/**
 * @since 2.10.0
 */
exports.bind = 
/*#__PURE__*/
Chain_1.bind(exports.Chain);
/**
 * @since 2.10.0
 */
exports.bindTo = 
/*#__PURE__*/
Functor_1.bindTo(exports.Functor);
/**
 * ixDo-notation combinator for discarding the result of a computation rather
 * than binding it to a name. It is simply an alias for `ixchainFirst`
 * @since 2.10.0
 */
exports.ixdiscard = exports.ixchainFirst;
/**
 * Do-notation combinator for discarding the result of a computation rather
 * than binding it to a name. It is simply an alias for `chainFirst`
 * @since 2.10.0
 */
exports.discard = exports.chainFirst;
// -------------------------------------------------------------------------------------
// pipeable sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 2.10.0
 */
exports.ixapS = 
/*#__PURE__*/
IxApply_1.ixapS(exports.IxApply);
/**
 * @since 2.10.0
 */
exports.apS = 
/*#__PURE__*/
Apply_1.apS(exports.Apply);
